(* logging *)

let log_channel = ref None 
let log_file = ref "" 

let log_or_print fmt = 
  let k result = 
    match !log_channel with
      None -> print_string result ; flush stdout
    | Some(o) -> output_string o result ; flush o
  in 
  Printf.kprintf k fmt 
    
let log_only fmt =
  let k result = 
    match !log_channel with
      None -> () 
    | Some(o) -> output_string o result ; flush o
  in 
  Printf.kprintf k fmt

let log_and_print fmt = 
  let k result = begin
    output_string stdout result ; flush stdout ;
    match !log_channel with
      None -> ()
    | Some(o) -> output_string o result ; flush o
  end 
  in 
  Printf.kprintf k fmt 

let my_read size fd buff name = 
  let sofar = ref 0 in
  while !sofar < size do 
    let this_chunk = Unix.read fd buff !sofar (size - !sofar) in
    if this_chunk = 0 then begin
      failwith (Printf.sprintf "read %d of %d bytes from [%s]"
        !sofar size name) 
    end else 
      sofar := !sofar + this_chunk
  done 

let load_file name =
  try begin
    Stats.time "loading files" (fun () -> 
    let stats = Unix.stat name in
    let size = stats.Unix.st_size in 
    if size = 0 then 
      log_and_print "WARNING: [%s] is a 0 byte file\n" name 
    else if size < 0 then begin
      log_and_print "ERROR: [%s] has reported size %d\n" name size ;
      failwith ("error loading " ^ name)
    end else if size > Sys.max_string_length then begin
      log_and_print "ERROR: [%s] has size %d: TOO BIG FOR ME (max %d)\n" 
          name size Sys.max_string_length;
      failwith ("error loading " ^ name)
    end ; 
    let buff = String.make size '\000' in
    let fd = Unix.openfile name [Unix.O_RDONLY] 0 in
    my_read size fd buff name ; 
    Unix.close fd ; 
    log_only "[%s] loaded, %d bytes\n" name size ; 
    buff) () 
  end
  with e -> 
    log_and_print "ERROR: error loading [%s]\n" name ;
    raise e

let init_log filename = 
  try
    let out = open_out filename in
    log_channel := Some(out) ;
    log_file := filename ;
    Array.iter (fun s -> log_only " %s" s) Sys.argv ;
    log_only "\n"  
  with e ->
    Printf.printf "WARNING: unable to open log file [%s]: %s"
      filename (Printexc.to_string e) ;
    () 


type loc = {
  l_line   : int ;
  l_col    : int ;
  l_offset : int ;
} 

type span = loc * loc 

let loc_unknown = { l_line = 0; l_col = 0; l_offset = 0; }

let span_unknown = (loc_unknown,loc_unknown)

let loc_to_str l = 
  Printf.sprintf "[Line %d, Col %d, Off %d]" l.l_line l.l_col l.l_offset

let span_to_str (l1,l2) = 
  Printf.sprintf "[Line %4d, Col %2d -- Line %4d, Col %2d]" 
    l1.l_line l1.l_col
    l2.l_line l2.l_col

(* for convenient parse errors *)

type input_context = {
  mutable loc : loc ;
  mutable filename : string ;
  mutable lexbuf : Lexing.lexbuf ; 
  mutable warn_only : bool ; 
} 
let the_context = {
  loc = loc_unknown ;
  filename = "" ;
  lexbuf = Lexing.from_string "";
  warn_only = false ;
} 

let lex_init (filename: string) (inchannel: in_channel) : Lexing.lexbuf =
  let lexbuf = Lexing.from_channel inchannel in
  the_context.loc <- { l_line = 1; l_col = 1; l_offset = 1; } ;
  the_context.filename <- filename ;
  the_context.lexbuf <- lexbuf ;
  lexbuf 

let newline () = 
  let c = the_context in 
  c.loc <- { l_line = c.loc.l_line + 1 ; 
             l_col = 1 ; 
             l_offset = c.loc.l_offset + 1 ; } 

let tab () = 
  let c = the_context in
  c.loc <- { c.loc with l_col = c.loc.l_col + 8 - (c.loc.l_col mod 8) ; 
                        l_offset = c.loc.l_offset + 1 } 

let adj (buf : string) (thing : 'a) : ('a) = 
  let c = the_context in 
  let start_loc = c.loc in
  let len = (String.length buf) in 
  let end_loc = { c.loc with l_col = c.loc.l_col + len - 1} in 
  c.loc <- { c.loc with l_col = c.loc.l_col + len } ;
  let span = (start_loc, end_loc) in 
  (thing) 

let strip str = 
  let len = String.length str in
  String.sub str 1 (len - 2) 

let input_error sort_msg msg =
  let c = the_context in 
  let near_text = Lexing.lexeme c.lexbuf in 
  log_and_print 
  "\n[%s] %s %s at line %d column %d\nNear Text: %s\n\t%s\n"
    c.filename 
    sort_msg
    (if c.warn_only then "WARNING" else "ERROR") 
    c.loc.l_line c.loc.l_col near_text msg ;
  raise Parsing.Parse_error

let lex_error msg = input_error "LEXER" msg 
let parse_error msg = input_error "PARSE" msg 
